from natural_selection.genetic_algorithms.operators.selection.elite_selection import *
from natural_selection.genetic_algorithms.operators.selection.survivor_selection import *
from natural_selection.genetic_algorithms.operators.selection.parent_selection import *